#include <abort.h>
#include <malloc.h>
#include "hash_.h"

Hash_* hash_New(size_t keysize, size_t valuesize, keys_equ_f keys_equ) {

	Hash_* hash = malloc(sizeof(Hash_));
	if(NULL == hash)
		abort();
	hash_Construct(hash, keysize, valuesize, keys_equ);
	return hash;
}

void hash_Delete(Hash_* hash) {

	hash_Destroy(hash);
	free(hash);
}

void hash_Construct(Hash_* hash, size_t keysize, size_t valuesize, keys_equ_f keys_equ) {

	hash->keysize = keysize;
	hash->valuesize = valuesize;
	hash->keys_equ = keys_equ;
	vec_Construct(&hash->pairs, sizeof(Pair_), 0);
}

void hash_Destroy(Hash_* hash) {

	for(uint32_t i=0; i<hash_GetNumKeys(hash); i++) {
		pair_Delete(hash_GetPair(hash, i));
	}
	vec_Destroy(&hash->pairs);
}

void hash_Insert(Hash_* hash, void* key, void* value) {

	Pair_* pair = NULL;
	for(uint32_t i=0, n=vec_GetSize(&hash->pairs); i<n; i++) {
		pair = vec_GetElement(&hash->pairs, i);
		if(hash->keys_equ(key, pair->key))
			break;
		pair = NULL;
	}

	if(NULL == pair) {
		/* didn't find desired key - insert it */
		Pair_* newpair = pair_New(hash->keysize, hash->valuesize, key);
		vec_PushBack(&hash->pairs, newpair);
		pair = newpair;
	}

	vec_PushBack(&pair->vals, value);
}

inline uint32_t hash_GetNumKeys(Hash_* hash) {

	return vec_GetSize(&hash->pairs);
}

inline Pair_* hash_GetPair(Hash_* hash, uint32_t i) {

	return vec_GetElement(&hash->pairs, i);
}

