#include <string.h>
#include <sys/region.h>
#include "stack.h"

void stackConstruct(Stack* stack, size_t el_size) {

	vec_Construct(&stack->data, el_size, ALIGN_DEFAULT);
	stack->el_size = el_size;
	stack->el_pushed = 0;
}

void stackDestroy(Stack* stack) {

	vec_Destroy(&stack->data);
}

void stackPush(Stack* stack, void* element) {

	if(vec_GetSize(&stack->data) <= stack->el_pushed) {
		vec_Resize(&stack->data, vec_GetSize(&stack->data) << 1);
	}

	void* top = vec_GetElement(&stack->data, stack->el_pushed++);
	memcpy(top, element, stack->el_size);
}

void* stackPop(Stack* stack) {

	if(0 == stack->el_pushed)
		return NULL;

	return vec_GetElement(&stack->data, --stack->el_pushed);
}

void* stackGet(Stack* stack, uint_t i) {

	if(i+1 > stack->el_pushed)
		return NULL;

	return vec_GetElement(&stack->data, stack->el_pushed - i - 1);
}

inline bool stackIsEmpty(const Stack* stack) {

	return 0 == stack->el_pushed;
}
