#include <abort.h>
#include <malloc.h>
#include <string.h>
#include "vec_.h"

Vec_* vec_New(size_t el_size, size_t alignment) {

	Vec_* vec = malloc(sizeof(Vec_));
	if(NULL == vec)
		abort();
	vec_Construct(vec, el_size, alignment);
	return vec;
}

inline void vec_Delete(Vec_* vec) {

	vec_Destroy(vec);
	free(vec);
}

void vec_Construct(Vec_* vec, size_t el_size, size_t alignment) {

	vec->el_size = el_size;
	vec->alignment = alignment;
	vec->num_el = 0;
	vec->max_el = 1;
	if( NULL == (vec->data_unaligned = malloc(el_size*vec->max_el + (1<<alignment)))) {
		abort();
	}
	vec->end = vec->data = ALIGN_ADR(vec->data_unaligned, alignment);
}

inline void vec_Destroy(Vec_* vec) {

	free(vec->data);
}

void vec_Resize(Vec_* vec, uint32_t newsize) {

	if( NULL == (vec->data_unaligned = realloc(vec->data_unaligned, vec->el_size*newsize + (1<<vec->alignment)))) {
		vec->data = NULL;
		vec->end = NULL;
		abort();
	}
	vec->data = ALIGN_ADR(vec->data_unaligned, vec->alignment);
	vec->end = (uint8_t*)vec->data + vec->el_size*vec->num_el;
	vec->max_el = newsize;
}

void vec_PushBack(Vec_* vec, const void* element) {

	if(vec->num_el >= vec->max_el)
		vec_Resize(vec, vec->max_el + 16);	// it used to be vec->max_el << 1, but seems like it triggers a BUG!

	memcpy(vec->end, element, vec->el_size);
	vec->end = (uint8_t*)vec->end + vec->el_size;
	vec->num_el++;
}

inline void* vec_GetElement(Vec_* vec, uint32_t i) {

	return (uint8_t*)vec->data + i*vec->el_size;
}

inline void* vec_GetTail(Vec_* vec) {

	return (uint8_t*)vec->data + vec->num_el*vec->el_size;
}

inline const void* vec_GetConstElement(const Vec_* vec, uint32_t i) {

	return (uint8_t*)vec->data + i*vec->el_size;
}

inline uint32_t vec_GetSize(const Vec_* vec) {

	return vec->num_el;
}

inline void vec_Reset(Vec_* vec) {

	vec->num_el = 0;
	vec->end = vec->data;
}

