#include <abort.h>
#include <malloc.h>
#include <stream.h>
#include <ultra64.h>
#include <sys/region.h>
#include <math/vector.h>
#include <math/math.h>
#include <sys/sys.h>
#include "dltrans.h"

#define VTXBUFFER_MAXELEMENTS	64
#define IDXBUFFER_MAXELEMENTS	64

static Lights1 mat1_light = gdSPDefLights1(0, 0, 0, 255, 255, 255, 0, 0, 127);

static Vp vp = {
	{
		{(BUFFER_W/2)*4, (BUFFER_H/2)*4, G_MAXZ, 0},
		{(BUFFER_W/2)*4, (BUFFER_H/2)*4, 0, 0}
	}
};

static Gfx rspinit_dl[] = {
	gsSPViewport(&vp),
	gsSPEndDisplayList()
};

static Gfx rdpinit_dl[] = {
	gsDPSetCycleType(G_CYC_1CYCLE),
	gsDPPipelineMode(G_PM_1PRIMITIVE),
	gsDPSetScissor(G_SC_NON_INTERLACE, 0, 0, BUFFER_W, BUFFER_H),
	gsDPSetTextureLOD(G_TL_TILE),
	gsDPSetTextureLUT(G_TT_NONE),
	gsDPSetTextureDetail(G_TD_CLAMP),
	gsDPSetTexturePersp(G_TP_PERSP),
	gsDPSetTextureFilter(G_TF_BILERP),
	gsDPSetTextureConvert(G_TC_FILT),
	gsDPSetCombineMode(G_CC_SHADE, G_CC_SHADE),
	gsDPSetCombineKey(G_CK_NONE),
	gsDPSetAlphaCompare(G_AC_NONE),
	gsDPSetRenderMode(G_RM_OPA_SURF, G_RM_OPA_SURF2),
	gsDPSetColorDither(G_CD_DISABLE),
	gsDPPipeSync(),
	gsSPEndDisplayList()
};

void dltransConstruct(DLTrans* dltrans) {

	vec_Construct(&dltrans->vtxBuffer, sizeof(Vtx), ALIGN_8B);
	vec_Resize(&dltrans->vtxBuffer, VTXBUFFER_MAXELEMENTS);
	dltrans->idxBuffer = malloc(sizeof(*dltrans->idxBuffer)*IDXBUFFER_MAXELEMENTS);
	dltransSetScale(dltrans, 1.0f);
}

void dltransDestroy(DLTrans* dltrans) {

	free(dltrans->idxBuffer);
	vec_Destroy(&dltrans->vtxBuffer);
}

inline void dltransSetScale(DLTrans* dltrans, float scale) {

	dltrans->scale = scale;
}

void dltransProcess(DLTrans* dltrans, const DList* dl_in, DisplayList* dl_out, size_t* dl_out_size, void** data_out, size_t* data_out_size) {

	if(dltrans) {}
	if(dl_in) {}
	if(dl_out) {}
	if(dl_out_size) {}
	if(data_out) {}
	if(data_out_size) {}

	DisplayList* rsp_dl = dl_out;

	Stream stream;
	streamConstruct(&stream, dlistGetData(dl_in), dlistGetSize(dl_in));

	name_t name;
	do {
		name = *(name_t*)streamGetData(&stream);
		switch(name) {
			case Init: {
				(void*)streamExtract(&stream, sizeof(Command_Init));

				dlAddCommand(rsp_dl, gsSPSegment(0x0, 0x0));
				dlAddCommand(rsp_dl, gsSPSegment(0x8, 0x0));
				dlAddCommand(rsp_dl, gsSPDisplayList(rspinit_dl));
				dlAddCommand(rsp_dl, gsSPDisplayList(rdpinit_dl));
				break;
			}
			case SetVertexSource: {
				Command_SetVertexSource* cmd = streamExtract(&stream, sizeof(Command_SetVertexSource));
				dltrans->vtxFormat = cmd->format;
				dltrans->vtxSource = cmd->data;
				break;
			}
			case SetIndexSource: {
				Command_SetIndexSource* cmd = streamExtract(&stream, sizeof(Command_SetIndexSource));
				dltrans->idxSource = cmd->data;
				break;
			}
			case LoadVertexBuffer: {
				Command_LoadVertexBuffer* cmd = streamExtract(&stream, sizeof(Command_LoadVertexBuffer));

				uint32_t n = cmd->nvts;
				if(n > VTXBUFFER_MAXELEMENTS) {
					abort();
				}

				Vertex* vtxSource = (Vertex*)dltrans->vtxSource + cmd->src_ofs;
				Vtx* vtx = vec_GetElement(&dltrans->vtxBuffer, cmd->dst_ofs);
				while(n--) {
					vtx->v.ob[0] = vtxSource->position.x * dltrans->scale;
					vtx->v.ob[1] = vtxSource->position.y * dltrans->scale;
					vtx->v.ob[2] = vtxSource->position.z * dltrans->scale;
					vtx->v.tc[0] = vtxSource->texcoords.x * 32 * textureGetWidth(&dltrans->tex);
					vtx->v.tc[1] = vtxSource->texcoords.y * 32 * textureGetHeight(&dltrans->tex);
					vtx->v.cn[0] = vtx->v.cn[1] = vtx->v.cn[2] = vtx->v.cn[3] = 0xFF;

					vtx++;
					vtxSource++;
				}

				break;
			}
			case LoadIndexBuffer: {
				Command_LoadIndexBuffer* cmd = streamExtract(&stream, sizeof(Command_LoadIndexBuffer));

				uint32_t n = cmd->nids;
				if(n > IDXBUFFER_MAXELEMENTS) {
					abort();
				}

				uint32_t* idxSource = (uint32_t*)dltrans->idxSource + cmd->src_ofs;
				uint32_t* idx = dltrans->idxBuffer + cmd->dst_ofs;
				while(n--) {
					*idx++ = *idxSource++;
				}

				break;
			}
			case DrawPrimitives: {
				Command_DrawPrimitives* cmd = streamExtract(&stream, sizeof(Command_DrawPrimitives));

				Vtx* vtx = vec_GetElement(&dltrans->vtxBuffer, 0);
				uint32_t* idx = dltrans->idxBuffer;
				if(PrimTriangle == cmd->type) {
					for(uint32_t i=0; i<cmd->num; i++) {
						dlAddCommand(rsp_dl, gsSPVertex(&vtx[*idx++], 1, 0));
						dlAddCommand(rsp_dl, gsSPVertex(&vtx[*idx++], 1, 1));
						dlAddCommand(rsp_dl, gsSPVertex(&vtx[*idx++], 1, 2));
						dlAddCommand(rsp_dl, gsSP1Triangle(0, 1, 2, 0));
					}
				}
				else {
					abort();
				}

				break;
			}
			case SetBuffer: {
				Command_SetBuffer* cmd = streamExtract(&stream, sizeof(Command_SetBuffer));
				Surface* surface = NULL;

				if(ColorBuffer == cmd->buftype) {
					surface = &dltrans->surfColor;
					dlAddCommand(rsp_dl, gsDPSetColorImage(G_IM_FMT_RGBA, G_IM_SIZ_16b, cmd->width, cmd->buf));
				} else if(ZBuffer == cmd->buftype) {
					surface = &dltrans->surfZ;
					dlAddCommand(rsp_dl, gsDPSetDepthImage(cmd->buf));
				} else {
					abort();
				}

				surface->pixfmt = cmd->format;
				surface->data = cmd->buf;
				surface->width = cmd->width;
				surface->height = cmd->height;

				break;
			}
			case ClearBuffers: {
				Command_ClearBuffers* cmd = streamExtract(&stream, sizeof(Command_ClearBuffers));
				uint32_t clrval;
				Surface* surface = NULL;

				if(cmd->types & ColorBuffer) {
					clrval = GPACK_RGBA5551((uint8_t)(255*dltrans->clearColor.x),
								(uint8_t)(255*dltrans->clearColor.y),
								(uint8_t)(255*dltrans->clearColor.z),
								1);
					clrval = clrval << 16 | clrval;
					surface = &dltrans->surfColor;

					dlAddCommand(rsp_dl, gsDPSetColorImage(G_IM_FMT_RGBA, G_IM_SIZ_16b, surface->width, surface->data));
					dlAddCommand(rsp_dl, gsDPSetCycleType(G_CYC_FILL));
					dlAddCommand(rsp_dl, gsDPSetRenderMode(G_RM_NOOP, G_RM_NOOP2));
					dlAddCommand(rsp_dl, gsDPSetFillColor(clrval));
					dlAddCommand(rsp_dl, gsDPFillRectangle(0, 0, surface->width - 1, surface->height - 1));
					dlAddCommand(rsp_dl, gsDPPipeSync());
				}
				if(cmd->types & ZBuffer) {
					clrval = GPACK_ZDZ((uint16_t)(G_MAXFBZ*dltrans->clearZVal), 0);
					clrval = clrval << 16 | clrval;
					surface = &dltrans->surfZ;

					dlAddCommand(rsp_dl, gsDPSetColorImage(G_IM_FMT_RGBA, G_IM_SIZ_16b, surface->width, surface->data));
					dlAddCommand(rsp_dl, gsDPSetCycleType(G_CYC_FILL));
					dlAddCommand(rsp_dl, gsDPSetRenderMode(G_RM_NOOP, G_RM_NOOP2));
					dlAddCommand(rsp_dl, gsDPSetFillColor(clrval));
					dlAddCommand(rsp_dl, gsDPFillRectangle(0, 0, surface->width - 1, surface->height - 1));
					dlAddCommand(rsp_dl, gsDPPipeSync());
				}

				break;
			}
			case SetClearColor: {
				Command_SetClearColor* cmd = streamExtract(&stream, sizeof(Command_SetClearColor));

				dltrans->clearColor.x = cmd->r;
				dltrans->clearColor.y = cmd->g;
				dltrans->clearColor.z = cmd->b;
				dltrans->clearColor.t = cmd->a;

				break;
			}
			case SetClearZValue: {
				Command_SetClearZValue* cmd = streamExtract(&stream, sizeof(Command_SetClearZValue));

				dltrans->clearZVal = cmd->val;

				break;
			}
			case LoadTexture: {
				Command_LoadTexture* cmd = streamExtract(&stream, sizeof(Command_LoadTexture));

				dltrans->tex = cmd->texture;
				if(TEXTURE_FMT_RGBA32 == cmd->texture.fmt) {
					dlAddCommand(rsp_dl, gsDPLoadTextureBlock(cmd->texture.pixels,
							G_IM_FMT_RGBA, G_IM_SIZ_32b, cmd->texture.width, cmd->texture.height,
							0, G_TX_WRAP, G_TX_WRAP, log2uint(cmd->texture.width), log2uint(cmd->texture.height),
							G_TX_NOLOD, G_TX_NOLOD));
					dlAddCommand(rsp_dl, gsDPSetTextureLUT(G_TT_NONE));
				} else if(TEXTURE_FMT_RGBA16 == cmd->texture.fmt) {
					dlAddCommand(rsp_dl, gsDPLoadTextureBlock(cmd->texture.pixels,
							G_IM_FMT_RGBA, G_IM_SIZ_16b, cmd->texture.width, cmd->texture.height,
							0, G_TX_WRAP, G_TX_WRAP, log2uint(cmd->texture.width), log2uint(cmd->texture.height),
							G_TX_NOLOD, G_TX_NOLOD));
					dlAddCommand(rsp_dl, gsDPSetTextureLUT(G_TT_NONE));
				} else if(TEXTURE_FMT_CI4 == cmd->texture.fmt) {
					dlAddCommand(rsp_dl, gsDPLoadTextureBlock_4b(cmd->texture.pixels,
							G_IM_FMT_CI, cmd->texture.width, cmd->texture.height,
							0, G_TX_WRAP, G_TX_WRAP, log2uint(cmd->texture.width), log2uint(cmd->texture.height),
							G_TX_NOLOD, G_TX_NOLOD));
					dlAddCommand(rsp_dl, gsDPLoadTLUT_pal16(0, cmd->texture.palette));
					dlAddCommand(rsp_dl, gsDPSetTextureLUT(G_TT_RGBA16));
				} else {
					abort();
				}
				break;
			}
			case Jump: {
				abort();
				Command_Jump* cmd = streamExtract(&stream, sizeof(Command_Jump));
				DisplayList* child_dl = displaylistNew();

				dltransProcess(dltrans, cmd->dl, child_dl, NULL, NULL, NULL);

				break;
			}
			case End: {
				(void*)streamExtract(&stream, sizeof(Command_End));

				dlAddCommand(rsp_dl, gsSPEndDisplayList());
				break;
			}
			case SetTextureScale: {
				Command_SetTextureScale* cmd = streamExtract(&stream, sizeof(Command_SetTextureScale));

				dlAddCommand(rsp_dl, gsSPTexture((s32)(0x8000*cmd->s), (s32)(0x8000*cmd->t), 0, G_TX_RENDERTILE, G_ON));

				// temporary stuff
				dlAddCommand(rsp_dl, gsSPSetGeometryMode(G_SHADE | G_SHADING_SMOOTH | G_CULL_BACK | G_ZBUFFER | G_LIGHTING));
				dlAddCommand(rsp_dl, gsSPSetLights1(mat1_light));
				dlAddCommand(rsp_dl, gsDPPipeSync());
				dlAddCommand(rsp_dl, gsDPSetCycleType(G_CYC_1CYCLE));
				dlAddCommand(rsp_dl, gsDPSetRenderMode(G_RM_ZB_OPA_SURF, G_RM_ZB_OPA_SURF2));

				break;
			}
			case LoadMatrix: {
				Command_LoadMatrix* cmd = streamExtract(&stream, sizeof(Command_LoadMatrix));
				guMtxF2L(*cmd->mtx, cmd->mtx_native);

				if(MATRIX_PROJECTION == cmd->type) {
					dlAddCommand(rsp_dl, gsSPMatrix(cmd->mtx_native, G_MTX_PROJECTION | G_MTX_LOAD | G_MTX_NOPUSH));
				} else if(MATRIX_MODELVIEW == cmd->type) {
					dlAddCommand(rsp_dl, gsSPMatrix(cmd->mtx_native, G_MTX_MODELVIEW | G_MTX_LOAD | G_MTX_NOPUSH));
				} else {
					abort();
				}
				break;
			}
			case LoadLight: {
				(void*)streamExtract(&stream, sizeof(Command_LoadLight));

				break;
			}
			case Sync: {
				(void*)streamExtract(&stream, sizeof(Command_Sync));

				dlAddCommand(rsp_dl, gsDPFullSync());
				break;
			}
			case JumpToTranslated: {
				Command_JumpToTranslated* cmd = streamExtract(&stream, sizeof(Command_JumpToTranslated));

				dlAddCommand(rsp_dl, gsSPDisplayList((Gfx*)cmd->dl));
				break;
			}
			case SetRenderMode: {
				Command_SetRenderMode* cmd = streamExtract(&stream, sizeof(Command_SetRenderMode));

				if(RM_SHADE == cmd->mode) {
					dlAddCommand(rsp_dl, gsDPSetCombineMode(G_CC_SHADE, G_CC_SHADE));
				} else if(RM_DECAL == cmd->mode) {
					dlAddCommand(rsp_dl, gsDPSetCombineMode(G_CC_DECALRGB, G_CC_DECALRGB));
				} else {
					abort();
				}
				break;
			}
			case NormalizePerspective: {
				Command_NormalizePerspective* cmd = streamExtract(&stream, sizeof(Command_NormalizePerspective));

				dlAddCommand(rsp_dl, gsSPPerspNormalize((u16)cmd->scale));
				break;
			}
			
			default: {
				abort();
				break;
			}
		}
	} while(End != name);

	streamDestroy(&stream);
}

