#ifndef ENGINE_H
#define ENGINE_H

#include <containers/vec_.h>
#include <math/matrix.h>
#include <math/vector.h>
#include <node/billboardtnode.h>
#include <node/cameratnode.h>
#include <node/luxtnode.h>
#include <node/meshtnode.h>
#include <node/object2d.h>
#include <sys/surface.h>
#include <sys/displaylist.h>
#include <sys/dlist.h>
#include <dltrans.h>
#include <viewport.h>

typedef struct {
	Vec_		lights;
	Viewport*	viewport;
	CameraTNode*	camera;
	const Surface*	renderTarget;
	const Surface*	zBuffer;
	Vec_		rspMatrixes;
	Mtx		projMtx;
	Matrix		projMtxF;
	DList		dl;
	DList		my_dlist;
	DLTrans		dltrans;
	DisplayList	rsp_cfb;
	DisplayList	rsp_zb;
} Engine;

#include "task.h"

void engineConstruct(Engine* engine);
void engineSetViewport(Engine* engine, Viewport* viewport);
void engineSetCamera(Engine* engine, CameraTNode* camera);
void engineSetRenderTarget(Engine* engine, const Surface* renderTarget);
void engineSetZBuffer(Engine* engine, const Surface* zbuffer);
Viewport* engineGetViewport(const Engine* engine);
CameraTNode* engineGetCamera(const Engine* engine);
const Surface* engineGetRenderTarget(const Engine* engine);
const Surface* engineGetZBuffer(const Engine* engine);
void engineAddLight(Engine* engine, const Lux* light);
void engineRemoveLight(Engine* engine, const Lux* light);
void engineProcessTask(Engine* engine, const Task* task);

void engineSceneTaskF(Engine* engine, TNode* root);
void engineOverlayTaskF(Engine* engine, const float* alpha);
void engineImageTaskF(Engine* engine, TaskImageData* image);
void engineClearColorbufF(Engine* engine, void* arg);
void engineClearZbufF(Engine* engine, void* arg);

//Vector2f engineNDC2DC(float x, float y);

#endif	/* ENGINE_H */

