#include <ultra64.h>
#include "matrix.h"

inline void matrixIdentity(Matrix* mtx) {

	guMtxIdentF(*mtx);
}

inline void matrixRotate(Matrix* mtx, float a, float x, float y, float z) {

	guRotateF(*mtx, a, x, y, z);
}

inline void matrixScale(Matrix* mtx, float x, float y, float z) {

	guScaleF(*mtx, x, y, z);
}

inline void matrixTranslate(Matrix* mtx, float x, float y, float z) {

	guTranslateF(*mtx, x, y, z);
}

void matrixInvert(Matrix* restrict mi, Matrix* restrict mtxout) {

	float det = MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 2, 2)*MATRIX_GET(*mi, 3, 3)
		  + MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 2, 3)*MATRIX_GET(*mi, 3, 1)
		  + MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 2, 1)*MATRIX_GET(*mi, 3, 2)
		  + MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 2, 3)*MATRIX_GET(*mi, 3, 2)
		  
		  + MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 2, 0)*MATRIX_GET(*mi, 3, 3)
		  + MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 2, 2)*MATRIX_GET(*mi, 3, 0)
		  + MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 2, 1)*MATRIX_GET(*mi, 3, 3)
		  + MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 2, 3)*MATRIX_GET(*mi, 3, 0)
		  
		  + MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 2, 0)*MATRIX_GET(*mi, 3, 1)
		  + MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 2, 2)*MATRIX_GET(*mi, 3, 1)
		  + MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 2, 0)*MATRIX_GET(*mi, 3, 2)
		  + MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 2, 1)*MATRIX_GET(*mi, 3, 0)
		  
		  - MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 2, 3)*MATRIX_GET(*mi, 3, 2)
		  - MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 2, 1)*MATRIX_GET(*mi, 3, 3)
		  - MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 2, 2)*MATRIX_GET(*mi, 3, 1)
		  - MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 2, 2)*MATRIX_GET(*mi, 3, 3)
		  
		  - MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 2, 3)*MATRIX_GET(*mi, 3, 0)
		  - MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 2, 0)*MATRIX_GET(*mi, 3, 2)
		  - MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 2, 3)*MATRIX_GET(*mi, 3, 1)
		  - MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 2, 0)*MATRIX_GET(*mi, 3, 3)
		  
		  - MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 2, 1)*MATRIX_GET(*mi, 3, 0)
		  - MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 2, 1)*MATRIX_GET(*mi, 3, 2)
		  - MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 2, 2)*MATRIX_GET(*mi, 3, 0)
		  - MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 2, 0)*MATRIX_GET(*mi, 3, 1);
	
	MATRIX_GET(*mtxout, 0, 0) = ( MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 2, 2)*MATRIX_GET(*mi, 3, 3)
			+ MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 2, 3)*MATRIX_GET(*mi, 3, 1)
			+ MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 2, 1)*MATRIX_GET(*mi, 3, 2)
			- MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 2, 3)*MATRIX_GET(*mi ,3, 2)
			- MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 2, 1)*MATRIX_GET(*mi, 3, 3)
			- MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 2, 2)*MATRIX_GET(*mi, 3, 1) ) / det;

	MATRIX_GET(*mtxout, 0, 1) = ( MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 2, 3)*MATRIX_GET(*mi, 3, 2)
			+ MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 2, 1)*MATRIX_GET(*mi, 3, 3)
			+ MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 2, 2)*MATRIX_GET(*mi, 3, 1)
			- MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 2, 2)*MATRIX_GET(*mi, 3, 3)
			- MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 2, 3)*MATRIX_GET(*mi, 3, 1)
			- MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 2, 1)*MATRIX_GET(*mi, 3, 2) ) / det;

	MATRIX_GET(*mtxout, 0, 2) = ( MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 3, 3)
			+ MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 3, 1)
			+ MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 3, 2)
			- MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 3, 2)
			- MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 3, 3)
			- MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 3, 1) ) / det;

	MATRIX_GET(*mtxout, 0, 3) = ( MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 2, 2)
			+ MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 2, 3)
			+ MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 2, 1)
			- MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 2, 3)
			- MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 2, 1)
			- MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 2, 2) ) / det;

	MATRIX_GET(*mtxout, 1, 0) = ( MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 2, 3)*MATRIX_GET(*mi, 3, 2)
			+ MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 2, 0)*MATRIX_GET(*mi, 3, 3)
			+ MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 2, 2)*MATRIX_GET(*mi, 3, 0)
			- MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 2, 2)*MATRIX_GET(*mi, 3, 3)
			- MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 2, 3)*MATRIX_GET(*mi, 3, 0)
			- MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 2, 0)*MATRIX_GET(*mi, 3, 2) ) / det;

	MATRIX_GET(*mtxout, 1, 1) = ( MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 2, 2)*MATRIX_GET(*mi, 3, 3)
			+ MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 2, 3)*MATRIX_GET(*mi, 3, 0)
			+ MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 2, 0)*MATRIX_GET(*mi, 3, 2)
			- MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 2, 3)*MATRIX_GET(*mi, 3, 2)
			- MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 2, 0)*MATRIX_GET(*mi, 3, 3)
			- MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 2, 2)*MATRIX_GET(*mi, 3, 0) ) / det;

	MATRIX_GET(*mtxout, 1, 2) = ( MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 3, 2)
			+ MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 3, 3)
			+ MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 3, 0)
			- MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 3, 3)
			- MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 3, 0)
			- MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 3, 2) ) / det;

	MATRIX_GET(*mtxout, 1, 3) = ( MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 2, 3)
			+ MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 2, 0)
			+ MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 2, 2)
			- MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 2, 2)
			- MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 2, 3)
			- MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 2, 0) ) / det;

	MATRIX_GET(*mtxout, 2, 0) = ( MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 2, 1)*MATRIX_GET(*mi, 3, 3)
			+ MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 2, 3)*MATRIX_GET(*mi, 3, 0)
			+ MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 2, 0)*MATRIX_GET(*mi, 3, 1)
			- MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 2, 3)*MATRIX_GET(*mi, 3, 1)
			- MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 2, 0)*MATRIX_GET(*mi, 3, 3)
			- MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 2, 1)*MATRIX_GET(*mi, 3, 0) ) / det;

	MATRIX_GET(*mtxout, 2, 1) = ( MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 2, 3)*MATRIX_GET(*mi, 3, 1)
			+ MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 2, 0)*MATRIX_GET(*mi, 3, 3)
			+ MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 2, 1)*MATRIX_GET(*mi, 3, 0)
			- MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 2, 1)*MATRIX_GET(*mi, 3, 3)
			- MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 2, 3)*MATRIX_GET(*mi, 3, 0)
			- MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 2, 0)*MATRIX_GET(*mi, 3, 1) ) / det;

	MATRIX_GET(*mtxout, 2, 2) = ( MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 3, 3)
			+ MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 3, 0)
			+ MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 3, 1)
			- MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 3, 1)
			- MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 3, 3)
			- MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 3, 0) ) / det;

	MATRIX_GET(*mtxout, 2, 3) = ( MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 2, 1)
			+ MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 2, 3)
			+ MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 2, 0)
			- MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 2, 3)
			- MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 2, 0)
			- MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 2, 1) ) / det;

	MATRIX_GET(*mtxout, 3, 0) = ( MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 2, 2)*MATRIX_GET(*mi, 3, 1)
			+ MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 2, 0)*MATRIX_GET(*mi, 3, 2)
			+ MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 2, 1)*MATRIX_GET(*mi, 3, 0)
			- MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 2, 1)*MATRIX_GET(*mi, 3, 2)
			- MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 2, 2)*MATRIX_GET(*mi, 3, 0)
			- MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 2, 0)*MATRIX_GET(*mi, 3, 1) ) / det;

	MATRIX_GET(*mtxout, 3, 1) = ( MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 2, 1)*MATRIX_GET(*mi, 3, 2)
			+ MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 2, 2)*MATRIX_GET(*mi, 3, 0)
			+ MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 2, 0)*MATRIX_GET(*mi, 3, 1)
			- MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 2, 2)*MATRIX_GET(*mi, 3, 1)
			- MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 2, 0)*MATRIX_GET(*mi, 3, 2)
			- MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 2, 1)*MATRIX_GET(*mi, 3, 0) ) / det;

	MATRIX_GET(*mtxout, 3, 2) = ( MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 3, 1)
			+ MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 3, 2)
			+ MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 3, 0)
			- MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 3, 2)
			- MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 3, 0)
			- MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 3, 1) ) / det;

	MATRIX_GET(*mtxout, 3, 3) = ( MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 2, 2)
			+ MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 2, 0)
			+ MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 2, 1)
			- MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 2, 1)
			- MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 2, 2)
			- MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 2, 0) ) / det;
}

inline void matrixMultiply(Matrix* out, Matrix* ma, Matrix* mb) {
	
	guMtxCatF(*ma, *mb, *out);
}

inline void matrixPerspective(Matrix* out, float* perspNorm, float fovy, float aspect, float near, float far) {
	
	u16 u16_norm;
	guPerspectiveF((float (*)[4])out, &u16_norm, fovy, aspect, near, far, 1.0f);
	*perspNorm = (float)u16_norm;
}

