#include <abort.h>
#include <string.h>
#include <types.h>
#include <sys/region.h>
#include "tnode.h"

static Region	g_MtxRegion;
static Region	g_ResultMtxFRegion;

static void tnodeRegionInit(void) {
	if(	regionConstruct(&g_MtxRegion, MATRIX_ALIGNMENT, MAX_NODES*sizeof(Matrix), sizeof(Matrix)) < 0
		|| regionConstruct(&g_ResultMtxFRegion, MATRIX_ALIGNMENT, MAX_NODES*sizeof(Matrix), sizeof(Matrix)) < 0
	  )
		abort();
}

void tnodeConstruct(TNode* tnode, TNode* parent) {

	static bool moduleInitialized = false;
	if(!moduleInitialized) {
		tnodeRegionInit();
		moduleInitialized = true;
	}

	tnode->Process = tnodeProcess;
	tnode->mtx = (Matrix*)regionAlloc(&g_MtxRegion);
	tnode->resultMtxF = (Matrix*)regionAlloc(&g_ResultMtxFRegion);
	if(NULL == tnode->mtx || NULL == tnode->resultMtxF) {

		abort();
	}

	nodeConstruct(tnodeGetNode(tnode));

	if(parent) {
		nodeAttach(tnodeGetNode(parent), tnodeGetNode(tnode));
	}

	tnodeReset(tnode);
}

inline void tnodeDestroy(TNode* tnode) {

	nodeDestroy(tnodeGetNode(tnode));
}

inline Node* tnodeGetNode(TNode* tnode) {

	return &tnode->node;
}

inline Matrix* tnodeGetMatrix(const TNode* tnode) {

	return tnode->mtx;
}

inline Matrix* tnodeGetTransformedMatrix(const TNode* node) {

	return node->resultMtxF;
}

inline void tnodeReset(TNode* tnode) {

	matrixIdentity(tnodeGetMatrix(tnode));
}

void tnodeProcess(TNode* tnode) {

	Node* node = tnodeGetNode(tnode);
	TNode* parent = (TNode*)nodeGetParent(node);

	if(parent) {
		tnodeTransform(tnode, tnodeGetTransformedMatrix(parent));
	}

	memcpy(tnodeGetTransformedMatrix(tnode), tnodeGetMatrix(tnode), sizeof(Matrix));
	tnodeReset(tnode);
}

inline void tnodeTransform(TNode* tnode, Matrix* mtx) {

	matrixMultiply(tnodeGetMatrix(tnode), tnodeGetMatrix(tnode), mtx);
}

