#include <abort.h>
#include "string.h"
#include "malloc.h"

#define HEAPS_NUM			17
#define HEAP16_ENTRIES		2
#define HEAP16_SIZE			((1<<16)*HEAP16_ENTRIES)	// 64k
#define HEAP15_ENTRIES		2
#define HEAP15_SIZE			((1<<15)*HEAP15_ENTRIES)				// 32k
#define HEAP14_ENTRIES		2
#define HEAP14_SIZE			((1<<14)*HEAP14_ENTRIES)				// 16k
#define HEAP13_ENTRIES		25
#define HEAP13_SIZE			((1<<13)*HEAP13_ENTRIES)			// 8k
#define HEAP12_ENTRIES		50
#define HEAP12_SIZE			((1<<12)*HEAP12_ENTRIES)			// 4k
#define HEAP11_ENTRIES		50
#define HEAP11_SIZE			((1<<11)*HEAP11_ENTRIES)			// 2k
#define HEAP10_ENTRIES		50
#define HEAP10_SIZE			((1<<10)*HEAP10_ENTRIES)			// 1k
#define HEAP9_ENTRIES		50
#define HEAP9_SIZE			((1<<9)*HEAP9_ENTRIES)				// 512
#define HEAP8_ENTRIES		50
#define HEAP8_SIZE			((1<<8)*HEAP8_ENTRIES)				// 256
#define HEAP7_ENTRIES		50
#define HEAP7_SIZE			((1<<7)*HEAP7_ENTRIES)				// 128
#define HEAP6_ENTRIES		50
#define HEAP6_SIZE			((1<<6)*HEAP6_ENTRIES)				// 64
#define HEAP5_ENTRIES		50
#define HEAP5_SIZE			((1<<5)*HEAP5_ENTRIES)				// 32
#define HEAP4_ENTRIES		50
#define HEAP4_SIZE			((1<<4)*HEAP4_ENTRIES)				// 16
#define HEAP3_ENTRIES		50
#define HEAP3_SIZE			((1<<3)*HEAP3_ENTRIES)				// 8
#define HEAP2_ENTRIES		50
#define HEAP2_SIZE			((1<<2)*HEAP2_ENTRIES)				// 4
#define HEAP1_ENTRIES		0
#define HEAP1_SIZE			0						// 2
#define HEAP0_ENTRIES		0
#define HEAP0_SIZE			0						// 1

#define HEAP_MAX_ALLOC_SIZE	(2<<16)
#define HEAP_MIN_ALLOC_SIZE	4
#define HEAP_MAX_ENTRIES	50
#define	HEAP_FREE_ENTRY		0xFFFFFFFF
#define ALIGNMENT_MASK		0x7	// 8 byte boundary

typedef struct {
	void*		base;
	uint32_t	numobj;
	uint32_t	objsize;
	uint32_t	ofs[HEAP_MAX_ENTRIES];
} Heap;

static void* g_heapbase;
static Heap g_heaps[HEAPS_NUM];
static uint8_t g_buffer[HEAP_MAX_ALLOC_SIZE];

void mallocInit(void* base, size_t size) {
	static uint32_t heapsize[HEAPS_NUM] = {HEAP0_SIZE, HEAP1_SIZE, HEAP2_SIZE, HEAP3_SIZE, HEAP4_SIZE, HEAP5_SIZE, HEAP6_SIZE, HEAP7_SIZE,
		HEAP8_SIZE, HEAP9_SIZE, HEAP10_SIZE, HEAP11_SIZE, HEAP12_SIZE, HEAP13_SIZE, HEAP14_SIZE, HEAP15_SIZE, HEAP16_SIZE};
	static uint32_t heapentries[HEAPS_NUM] = {HEAP0_ENTRIES, HEAP1_ENTRIES, HEAP2_ENTRIES, HEAP3_ENTRIES, HEAP4_ENTRIES, HEAP5_ENTRIES, HEAP6_ENTRIES,
		HEAP7_ENTRIES, HEAP8_ENTRIES, HEAP9_ENTRIES, HEAP10_ENTRIES, HEAP11_ENTRIES, HEAP12_ENTRIES, HEAP13_ENTRIES, HEAP14_ENTRIES, HEAP15_ENTRIES,
		HEAP16_ENTRIES };
	uint32_t i,j;
	g_heapbase = base;
	
	for(j=0; j<HEAPS_NUM; base = (void*)((uintptr_t)base + heapsize[j++])) {
		g_heaps[j].base = base;
		g_heaps[j].numobj = heapentries[j];
		g_heaps[j].objsize = (1<<j);
		for(i=0; i<HEAP_MAX_ENTRIES; i++)
			g_heaps[j].ofs[i] = HEAP_FREE_ENTRY;
	}

	while((uintptr_t)g_heapbase + size < (uintptr_t)base) {}	// memory exceeded
}

void* realloc(void* ptr, size_t size) {
	if(!ptr) {
		return NULL;
	}
	
	memcpy(g_buffer, ptr, size);
	free(ptr);
	ptr = malloc(size);
	memcpy(ptr, g_buffer, size);
	
	return ptr;
}

void* malloc(size_t size) {
	size_t	heap, mask;
	u32 i;
	
	if(size>HEAP_MAX_ALLOC_SIZE)
		return NULL;
	if(size<HEAP_MIN_ALLOC_SIZE)
		size = HEAP_MIN_ALLOC_SIZE;
	
	mask = 0x80000000;
	heap = 31;
	while(!(size&mask)) {
		mask >>= 1;
		heap--;
	}
	if(size&(mask-1))
		heap++;
	
	for(i=0; i<g_heaps[heap].numobj; i++) {
		if(HEAP_FREE_ENTRY == g_heaps[heap].ofs[i]) {
			g_heaps[heap].ofs[i] = g_heaps[heap].objsize*i;
			uintptr_t p = (uintptr_t)g_heaps[heap].base + g_heaps[heap].ofs[i];
			if(p & ALIGNMENT_MASK) {
				abort();
			}
			return (void*)p;
		}
	}
	
	return NULL;
}

void free(void* ptr) {
	u32 i,j;
	
	for(j=0; j<HEAPS_NUM; j++) {
		for(i=0; i<g_heaps[j].numobj; i++) {
			if(((uintptr_t)g_heaps[j].base + g_heaps[j].ofs[i]) == (uintptr_t)ptr) {
				g_heaps[j].ofs[i] = HEAP_FREE_ENTRY;
				return;
			}
		}
	}
}

#if 0
void* malloc(size_t size) {
	if(size) {}
	return NULL;
}

void free(void* ptr) {
	if(ptr) {}
}

void* realloc(void* ptr, size_t size) {
	if(ptr && size) {}
	return NULL;
}
#endif

