#include <abort.h>
#include <malloc.h>
#include "dlist.h"

void dlistConstruct(DList* dlist) {

	dlist->size = 8192;
	dlist->data = malloc(dlist->size);
	dlistReset(dlist);
}

void dlistDestroy(DList* dlist) {
	free(dlist->data);
}

inline void dlistReset(DList* dlist) {

	dlist->current = dlist->data;
}

inline void* dlistGetData(const DList* dlist) {
	return dlist->data;
}

inline size_t dlistGetSize(const DList* dlist) {
	return dlist->size;
}

#define FOO(dl, T)			\
	T* p = (T*)((dl)->current);	\
	++(*(T**)(&(dl)->current));	\
	if((uint8_t*)dl->current >= (uint8_t*)dl->data + dl->size)	abort()

inline void dlistInit(DList* dlist) {

	FOO(dlist, Command_Init);
	p->name = Init;
}

inline void dlistSetVertexSource(DList* dlist, void* vts, uint8_t format) {

	FOO(dlist, Command_SetVertexSource);
	p->name = SetVertexSource;
	p->format = format;
	p->data = vts;
}

inline void dlistSetIndexSource(DList* dlist, void* ids) {

	FOO(dlist, Command_SetIndexSource);
	p->name = SetIndexSource;
	p->data = ids;
}

inline void dlistLoadVertexBuffer(DList* dlist, uint32_t src_ofs, uint32_t dst_ofs, uint32_t nvts) {

	FOO(dlist, Command_LoadVertexBuffer);
	p->name = LoadVertexBuffer;
	p->src_ofs = src_ofs;
	p->dst_ofs = dst_ofs;
	p->nvts = nvts;
}

inline void dlistLoadIndexBuffer(DList* dlist, uint32_t src_ofs, uint32_t dst_ofs, uint32_t nids) {

	FOO(dlist, Command_LoadIndexBuffer);
	p->name = LoadIndexBuffer;
	p->src_ofs = src_ofs;
	p->dst_ofs = dst_ofs;
	p->nids = nids;
}

inline void dlistDrawPrimitives(DList* dlist, uint8_t type, uint32_t num) {

	FOO(dlist, Command_DrawPrimitives);
	p->name = DrawPrimitives;
	p->type = type;
	p->num = num;
}

inline void dlistSetBuffer(DList* dlist, uint8_t buftype, uint8_t format, uint16_t width, uint16_t height, void* buf) {

	FOO(dlist, Command_SetBuffer);
	p->name = SetBuffer;
	p->buftype = buftype;
	p->format = format;
	p->width = width;
	p->height = height;
	p->buf = buf;
}

inline void dlistClearBuffers(DList* dlist, uint8_t types) {

	FOO(dlist, Command_ClearBuffers);
	p->name = ClearBuffers;
	p->types = types;
}

inline void dlistSetClearColor(DList* dlist, float r, float g, float b, float a) {

	FOO(dlist, Command_SetClearColor);
	p->name = SetClearColor;
	p->r = r;
	p->g = g;
	p->b = b;
	p->a = a;
}

inline void dlistSetClearZValue(DList* dlist, float val) {

	FOO(dlist, Command_SetClearZValue);
	p->name = SetClearZValue;
	p->val = val;
}

inline void dlistLoadTexture(DList* dlist, const Texture* texture) {

	FOO(dlist, Command_LoadTexture);
	p->name = LoadTexture;
	p->texture = *texture;
}

inline void dlistJump(DList* dlist, DList* dl) {

	FOO(dlist, Command_Jump);
	p->name = Jump;
	p->dl = dl;
}

inline void dlistEnd(DList* dlist) {

	FOO(dlist, Command_End);
	p->name = End;
}

inline void dlistSetTextureScale(DList* dlist, float s, float t) {

	FOO(dlist, Command_SetTextureScale);
	p->name = SetTextureScale;
	p->s = s;
	p->t = t;
}

inline void dlistLoadMatrix(DList* dlist, Matrix* mtx, Mtx* mtx_native, MatrixType type) {

	FOO(dlist, Command_LoadMatrix);
	p->name = LoadMatrix;
	p->mtx = mtx;
	p->mtx_native = mtx_native;
	p->type = type;
}

inline void dlistLoadLight(DList* dlist, void* light, uint16_t idx) {

	FOO(dlist, Command_LoadLight);
	p->name = LoadLight;
	p->light = light;
	p->idx = idx;
}

inline void dlistSync(DList* dlist) {

	FOO(dlist, Command_Sync);
	p->name = Sync;
}

inline void dlistJumpToTranslated(DList* dlist, void* dl) {

	FOO(dlist, Command_JumpToTranslated);
	p->name = JumpToTranslated;
	p->dl = dl;
}

inline void dlistSetRenderMode(DList* dlist, RenderMode mode) {

	FOO(dlist, Command_SetRenderMode);
	p->name = SetRenderMode;
	p->mode = mode;
}

inline void dlistNormalizePerspective(DList* dlist, float scale) {

	FOO(dlist, Command_NormalizePerspective);
	p->name = NormalizePerspective;
	p->scale = scale;
}

