#ifndef SYS_H
#define SYS_H

#include <types.h>
#include "displaylist.h"
#include "dlist.h"
#include "surface.h"
#include "time.h"

#define TMEM_SIZE	4096
#define	BUFFER_W	320			/*  */
#define	BUFFER_H	240			/*  */

#define STACKSIZE	0x2000
#define	NUM_PI_MSGS	8			/*  */
#define	MAINTHREAD_ID	3			/*  */
#define	MAINTHREAD_PRI	10			/*  */
#define	VBITHREAD_ID	4			/*  */
#define	VBITHREAD_PRI	12			/*  */
#define	SITHREAD_ID	5			/*  */
#define	SITHREAD_PRI	11			/*  */
#define AITHREAD_ID	6
#define AITHREAD_PRI	13

typedef enum {
	BUFFER_COLOR_DRAW,
	BUFFER_COLOR_DISPLAY,
	BUFFER_COLOR_0,
	BUFFER_COLOR_1,
	BUFFER_Z,
	BUFFER_AUX
} BufferType;

typedef enum {
	HANDLER_VBI,
	HANDLER_CONTROLLER
} HandlerType;

int		sysInit(void);
void		sysWaitForGPU(void);
void		sysWaitVBI(void);
void		sysRunGPU(const DList* dlist);
void		sysSwapBuffers(void);
const Surface*	sysGetBuffer(BufferType type);
DisplayList*	sysGetDisplayList(void);
void		sysError(void);
void		sysRegisterHandler(HandlerType type, void (*handler)(void*));
bool		sysIsRendering(void);
void*		sysGetHiMem(void);
Time		sysGetTime(void);
void		sysSetAudioSource(const void* src);
void 		sysSetAudioFrequency(uint32_t frequency);

#endif

