#ifndef TEXTURE_H
#define TEXTURE_H

#include <types.h>

typedef enum {
	TEXTURE_FMT_CI4,
	TEXTURE_FMT_RGBA16,
	TEXTURE_FMT_RGBA32
} TextureFormat;

typedef struct {
	uint16_t	width;
	uint16_t	height;
	void*		pixels;
	void*		palette;
	TextureFormat	fmt;
} Texture;

/* static initialization */
#define TEXTURE_INIT(w, h, pix, pal, f)	\
{ .width = w, .height = h, .pixels = pix, .palette = pal, .fmt = f }
/* interface */
void textureCopy(Texture* texture, const Texture* source);
void textureDestroy(Texture* texture);
uint32_t textureGetPixelCount(const Texture* texture);
uint16_t textureGetWidth(const Texture* texture);
uint16_t textureGetHeight(const Texture* texture);
uint8_t textureGetBytesPerPixel(const Texture* texture);
TextureFormat textureGetFormat(const Texture* texture);
void* textureGetPixels(Texture* texture);
const void* textureGetConstPixels(const Texture* texture);

#endif	/* TEXTURE_H */

