#include <actmgr/actmgr.h>
#include <interpol/interpol.h>
#include <math/math.h>
#include <postproc/postproc.h>
#include <resmgr/resmgr.h>
#include <sys/sys.h>
#include <abort.h>
#include <engine.h>
#include <stdlib.h>

static Engine	engine;
static float	alpha_overlay;
static float	angle, angle_torus;				/* in degrees */
static float	player_y = 160.0f;
static float	player_x = 500.0f, player_z = 200.0f;
static float	dir_x, dir_y, dir_z;
static uint32_t	framecnt;

static Material cube_mat = { {1.0f, 0.4f, 0.5f}, NULL };
static Pair cube_pair[] = {
	{&cube_mat, NULL },
	{NULL, NULL}
};
static DLTrans dltrans;
static DList cube_dl;
static DisplayList cube_translated_dl;
static Vertex cube_vts[] = {
	{{-0.5f, -0.5f, 0.5f}, {0.0f, 0.0f}},
	{{0.5f, -0.5f, 0.5f}, {0.0f, 0.0f}},
	{{0.5f, 0.5f, 0.5f}, {0.0f, 0.0f}},
	{{-0.5f, 0.5f, 0.5f}, {0.0f, 0.0f}},
	{{-0.5f, -0.5f, -0.5f}, {0.0f, 0.0f}},
	{{0.5f, -0.5f, -0.5f}, {0.0f, 0.0f}},
	{{0.5f, 0.5f, -0.5f}, {0.0f, 0.0f}},
	{{-0.5f, 0.5f, -0.5f}, {0.0f, 0.0f}},
};
static uint32_t cube_ids[] = {
	3, 2, 6,
	3, 6, 7,
	0, 1, 2,
	0, 2, 3,
	4, 5, 1,
	4, 1, 0,
	7, 6, 5,
	7, 5, 4,
	4, 0, 3,
	4, 3, 7,
	1, 5, 6,
	1, 6, 2,
};
static void generate_stuff(void) {
	
	dlistConstruct(&cube_dl);
	dlistSetVertexSource(&cube_dl, cube_vts, 0);
	dlistSetIndexSource(&cube_dl, cube_ids);
	dlistLoadVertexBuffer(&cube_dl, 0, 0, 8);
	dlistLoadIndexBuffer(&cube_dl, 0, 0, 3*12);
	dlistDrawPrimitives(&cube_dl, PrimTriangle, 12);
	dlistEnd(&cube_dl);

	dltransConstruct(&dltrans);
	displaylistConstruct(&cube_translated_dl);
	dltransSetScale(&dltrans, 20.0f);
	dltransProcess(&dltrans, &cube_dl, &cube_translated_dl, NULL, NULL, NULL);
	dltransSetScale(&dltrans, 1.0f);

	cube_pair[0].gfx = dlGetHead(&cube_translated_dl);
}

void HandlerVBI(void* arg) {

	if(arg) {}
	angle_torus += 1.0f;
	alpha_overlay += 1.0f/60;
	alpha_overlay = 0.25f*fabsf(sinf(2*M_PI*0.5f*framecnt/60.0f));
	framecnt++;
}

static void HandleController(void* arg) {

	static float sty_old;
	static float stx_old;
	static float ud_old;
	if(arg) {
		OSContPad* rdata = arg;
		angle -= 0.04f*stx_old + rdata->stick_x * 0.015f;

		dir_x = cosf(M_PI*(angle+90)/180);
		dir_y = dir_y;
		dir_z = -sinf(M_PI*(angle+90)/180);

		player_x += 0.0165*(rdata->stick_y + sty_old)*dir_x;
		player_z += 0.0165*(rdata->stick_y + sty_old)*dir_z;
		player_y += 0.03f*ud_old + (rdata->button & U_JPAD && TRUE) - (rdata->button & D_JPAD && TRUE);

		sty_old = sty_old*0.85f + 0.29f*rdata->stick_y;
		stx_old = 0.87f*stx_old + 0.012f*rdata->stick_x;
		ud_old = ud_old*0.904f + (rdata->button & U_JPAD && TRUE) - (rdata->button & D_JPAD && TRUE);
	}
}

static void ProcessTask(const Task* task) {

	if(sysIsRendering()) {
		sysWaitForGPU();
	}

	engineProcessTask(&engine, task);

	sysRunGPU(&engine.dl);
}

static void HandleSceneAction(Time time, void* arg) {

	if(arg) {}
	if(time.sec) {}

	static BillboardTNode	billboards[20];
	static float		randoms[20*3];
	static Viewport		viewport;
	static TNode		root;
	static CameraTNode	camera;
	static MeshTNode	mesh0;
	static MeshTNode	mesh1;
	static MeshTNode	mesh2;
	static LuxTNode		light0;
	static Matrix 		mv;
	static Task 		clearcfbTask = { (taskfptr)engineClearColorbufF, NULL };
	static Task 		clearzbufTask = { (taskfptr)engineClearZbufF, NULL };
	static Task 		sceneTask = { (taskfptr)engineSceneTaskF, &root };
	static bool 		initialized = false;

	if(!initialized) {
		generate_stuff();

		viewportConstruct(&viewport, BUFFER_W, BUFFER_H);
		tnodeConstruct(&root, NULL);
		luxtnodeConstruct(&light0);
		luxSetColor(luxtnodeGetLux(&light0), 0.2f, 0.2f, 1.0f);

		cameratnodeConstruct(&camera);
		cameraSetParams(cameratnodeGetCamera(&camera), 63.74f, 1.33f, 1000.0f);

		meshtnodeConstruct(&mesh0, resmgrGetResource("scene"));
		meshtnodeConstruct(&mesh1, resmgrGetResource("torus"));
		meshtnodeConstruct(&mesh2, resmgrGetResource("torus"));
		
		for(uint32_t i=0; i<sizeof(billboards)/sizeof(BillboardTNode); i++) {
			billboardtnodeConstruct(&billboards[i], resmgrGetResource("plane"));
			nodeAttach((Node*)&mesh2, (Node*)&billboards[i]);
			randoms[i*3+0] = 0.004f*rand()*(rand() > 0x4000 ? 1 : -1);
			randoms[i*3+1] = 0.001f*rand()*(rand() > 0x4000 ? 1 : -1);
			randoms[i*3+2] = 0.004f*rand()*(rand() > 0x4000 ? 1 : -1);
		}
		nodeAttach((Node*)&root, (Node*)&mesh0);
		nodeAttach((Node*)&root, (Node*)&mesh1);
		nodeAttach((Node*)&root, (Node*)&light0);
		nodeAttach((Node*)&root, (Node*)&camera);
		nodeAttach((Node*)&mesh1, (Node*)&mesh2);

		engineSetViewport(&engine, &viewport);
		engineSetCamera(&engine, &camera);
		engineSetZBuffer(&engine, sysGetBuffer(BUFFER_Z));

		sysRegisterHandler(HANDLER_CONTROLLER, HandleController);

		initialized = true;
	}
	/* camera */
	matrixRotate(&mv, angle, 0, 1, 0);
	tnodeTransform((TNode*)&camera, &mv);
	matrixTranslate(&mv, player_x, player_y, player_z);
	tnodeTransform((TNode*)&camera, &mv);
	/* billboards */
	for(uint32_t i=0; i<sizeof(billboards)/sizeof(BillboardTNode); i++) {
		matrixScale(&mv, 0.15f, 0.15f, 0.15f);
		tnodeTransform((TNode*)&billboards[i], &mv);
		matrixTranslate(&mv, randoms[i*3+0], randoms[i*3+1], -randoms[i*3+2]);
		tnodeTransform((TNode*)&billboards[i], &mv);
	}
	/* mesh 1 */
	matrixRotate(&mv, 1*angle_torus, 1, 1, 1);
	tnodeTransform((TNode*)&mesh1, &mv);
	matrixTranslate(&mv, 550, 260, 0);
	tnodeTransform((TNode*)&mesh1, &mv);
	/* mesh 2 */
	Vector3f pos0 = VECTOR3F_SET(0.0f, -30.0f, 0.0f);
	Vector3f pos1 = VECTOR3F_SET(0.0f, 30.0f, 0.0f);
	Vector3f pos = lerp3fCalculate(&pos0, &pos1, sinf(3.2*time.msec/1000.f));
	matrixScale(&mv, 0.4f, 0.4f, 0.4f);
	tnodeTransform((TNode*)&mesh2, &mv);
	matrixRotate(&mv, angle_torus, 0, 1, 0);
	tnodeTransform((TNode*)&mesh2, &mv);
	matrixTranslate(&mv, pos.x, pos.y, pos.z);
	tnodeTransform((TNode*)&mesh2, &mv);

	engineSetRenderTarget(&engine, sysGetBuffer(BUFFER_COLOR_DRAW));
	
	ProcessTask(&clearcfbTask);
	ProcessTask(&clearzbufTask);
	ProcessTask(&sceneTask);
}

static void HandleLogoAction(Time time, void* arg) {

	if(arg) {}
	if(time.sec) {}

	static Task imageTask = { (taskfptr)engineOverlayTaskF, NULL };
	ProcessTask(&imageTask);
}

void mainproc(void* arg) {
	if(arg) {}

	sysSetAudioFrequency(44100/2);
	sysSetAudioSource((void*)((uintptr_t)resmgrGetResource("sound") + 44));
	engineConstruct(&engine);
	sysRegisterHandler(HANDLER_VBI, HandlerVBI);

	/* actions */
	ActMgr actmgr;
	Action sceneAction;
	Action logoAction;
	Time tss = {0,1,0};
	Time tsd = {0,0,0};
	Time tls = {0,100,0};
	Time tld = {1,0,0};

	actionConstruct(&sceneAction, HandleSceneAction, NULL, tss, tsd);
	actionConstruct(&logoAction, HandleLogoAction, NULL, tls, tld);
	
	actmgrConstruct(&actmgr);
	actmgrAddAction(&actmgr, &sceneAction);
	actmgrAddAction(&actmgr, &logoAction);

	while(1) {
		actmgrProcess(&actmgr, sysGetTime());
		//const Surface* src = sysGetBuffer(BUFFER_AUX);
		//const Surface* dest = sysGetBuffer(BUFFER_COLOR_DRAW);
		sysWaitForGPU();
		//postprocBlur(dest, src);
		sysSwapBuffers();
	}
}

